'use strict';

var EventEmitter = require('events');
var util = require('util');
var getBrowserRTC = require('get-browser-rtc');
var once = require('once');
var Peer = require('./peer.js');
var transports = require('./transports.js');
try {
  var net = require('net');
} catch (err) {}
var remove = require('./util.js').remove;

module.exports = Exchange;
function Exchange(id, opts) {
  if (!id || typeof id !== 'string') {
    throw new Error('A network id must be specified');
  }
  if (!(this instanceof Exchange)) {
    return new Exchange(id, opts);
  }
  EventEmitter.call(this);

  opts = opts || {};
  var wrtc = opts.wrtc || getBrowserRTC();

  this._transports = opts.transports;
  if (!opts.transports) {
    this._transports = { websocket: transports.websocket };
    if (wrtc) {
      this._transports.webrtc = transports.webrtc({ wrtc: wrtc });
    }
    if (net) {
      this._transports.tcp = transports.tcp;
    }
  }

  this.id = id;
  this.peers = [];
  this._accepts = {};

  // lists of peers accepting connections, indexed by transport
  this._acceptPeers = {};
  for (var transportId in this._transports) {
    this._acceptPeers[transportId] = [];
  }
}
util.inherits(Exchange, EventEmitter);

Exchange.prototype.getNewPeer = function (cb) {
  var _this = this;

  cb = cb || function (err) {
    if (err) _this._error(err);
  };
  if (this.peers.length === 0) {
    return cb(new Error('Not connected to any peers. Connect to some seeds manually.'));
  }
  var peer = this.peers[Math.floor(Math.random() * this.peers.length)];
  peer.getNewPeer(cb);
};

Exchange.prototype.connect = function (transportId, address, opts, cb) {
  var _this2 = this;

  cb = once(cb || function (err) {
    if (err) _this2._error(err);
  });
  transportId = transportId.toLowerCase();
  if (!this._transports[transportId]) {
    return cb(new Error('Transport "' + transportId + '" not found'));
  }
  var connect = this._transports[transportId].connect;
  connect(address, opts, null, once(function (err, socket) {
    if (err) return cb(err);
    socket.transport = transportId;
    var peer = _this2._onConnection(socket, true);
    peer.once('error', cb);
    peer._onReady(function () {
      return cb(null, peer);
    });
  }));
};

Exchange.prototype._createPeer = function (socket, outgoing) {
  var _this3 = this;

  var peer = new Peer(this);
  peer.incoming = !outgoing;
  peer.once('error', function (err) {
    _this3.removePeer(peer);
    _this3.emit('peerError', err, peer);
  });
  peer.once('close', function () {
    _this3.removePeer(peer);
    _this3.emit('disconnect', peer);
  });
  peer._connect(socket);
  return peer;
};

Exchange.prototype.accept = function (transportId, opts, cb) {
  var _this4 = this;

  if (typeof opts === 'function') {
    cb = opts;
    opts = {};
  }
  cb = cb || function (err) {
    if (err) _this4._error(err);
  };
  transportId = transportId.toLowerCase();
  if (!this._transports[transportId]) {
    return cb(new Error('Transport "' + transportId + '" not found'));
  }
  var alreadyAcceptingError = new Error('Already accepting with "' + transportId + '" transport');
  if (this._accepts[transportId]) return cb(alreadyAcceptingError);
  var transport = this._transports[transportId];
  var register = function register(unaccept) {
    if (_this4._accepts[transportId]) {
      unaccept();
      return cb(alreadyAcceptingError);
    }

    _this4._accepts[transportId] = { opts: opts, unaccept: unaccept };
    var _iteratorNormalCompletion = true;
    var _didIteratorError = false;
    var _iteratorError = undefined;

    try {
      for (var _iterator = _this4.peers[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
        var peer = _step.value;

        peer._sendAccept(transportId, opts);
      }
    } catch (err) {
      _didIteratorError = true;
      _iteratorError = err;
    } finally {
      try {
        if (!_iteratorNormalCompletion && _iterator.return) {
          _iterator.return();
        }
      } finally {
        if (_didIteratorError) {
          throw _iteratorError;
        }
      }
    }

    cb(null);
  };
  if (transport.accept) {
    var onConnection = function onConnection(socket) {
      socket.transport = transportId;
      _this4._onConnection(socket);
    };
    transport.accept(opts, onConnection, function (err, unaccept) {
      if (err) return cb(err);
      if (typeof unaccept !== 'function') {
        return cb(new Error('Transport\'s "accept" function must return a cleanup function'));
      }
      register(unaccept);
    });
  } else {
    register();
  }
};

Exchange.prototype._onConnection = function (socket, outgoing) {
  var peer = this._createPeer(socket, outgoing);
  this.addPeer(peer);
  return peer;
};

Exchange.prototype.unaccept = function (transport, cb) {
  var _iteratorNormalCompletion2 = true;
  var _didIteratorError2 = false;
  var _iteratorError2 = undefined;

  try {
    for (var _iterator2 = this.peers[Symbol.iterator](), _step2; !(_iteratorNormalCompletion2 = (_step2 = _iterator2.next()).done); _iteratorNormalCompletion2 = true) {
      var peer = _step2.value;

      peer._sendUnaccept(transport);
    }
  } catch (err) {
    _didIteratorError2 = true;
    _iteratorError2 = err;
  } finally {
    try {
      if (!_iteratorNormalCompletion2 && _iterator2.return) {
        _iterator2.return();
      }
    } finally {
      if (_didIteratorError2) {
        throw _iteratorError2;
      }
    }
  }

  var unaccept = this._accepts[transport].unaccept;
  delete this._accepts[transport];
  if (unaccept) unaccept();
  if (cb) cb();
};

Exchange.prototype.addPeer = function (peer) {
  var _this5 = this;

  peer._onReady(function () {
    _this5.peers.push(peer);
    _this5.emit('peer', peer);
  });
};

Exchange.prototype.removePeer = function (peer) {
  peer.destroy();
  if (peer._accepts) {
    var _iteratorNormalCompletion3 = true;
    var _didIteratorError3 = false;
    var _iteratorError3 = undefined;

    try {
      for (var _iterator3 = Object.keys(peer._accepts)[Symbol.iterator](), _step3; !(_iteratorNormalCompletion3 = (_step3 = _iterator3.next()).done); _iteratorNormalCompletion3 = true) {
        var transportId = _step3.value;

        if (!this._acceptPeers[transportId]) continue;
        remove(this._acceptPeers[transportId], peer);
      }
    } catch (err) {
      _didIteratorError3 = true;
      _iteratorError3 = err;
    } finally {
      try {
        if (!_iteratorNormalCompletion3 && _iterator3.return) {
          _iterator3.return();
        }
      } finally {
        if (_didIteratorError3) {
          throw _iteratorError3;
        }
      }
    }
  }
  return remove(this.peers, peer);
};

Exchange.prototype.close = function (cb) {
  cb = cb || function () {};
  var _iteratorNormalCompletion4 = true;
  var _didIteratorError4 = false;
  var _iteratorError4 = undefined;

  try {
    for (var _iterator4 = this.peers[Symbol.iterator](), _step4; !(_iteratorNormalCompletion4 = (_step4 = _iterator4.next()).done); _iteratorNormalCompletion4 = true) {
      var peer = _step4.value;

      peer.destroy();
    }
  } catch (err) {
    _didIteratorError4 = true;
    _iteratorError4 = err;
  } finally {
    try {
      if (!_iteratorNormalCompletion4 && _iterator4.return) {
        _iterator4.return();
      }
    } finally {
      if (_didIteratorError4) {
        throw _iteratorError4;
      }
    }
  }

  for (var transportId in this._accepts) {
    this.unaccept(transportId);
  }
  cb(null);
};

Exchange.prototype._error = function (err) {
  this.emit('error', err);
};